/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  actingElement.cpp
///  \brief Implements the acting group elements.
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   07.10.2010
///
///  Copyright (C) 2010 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "actingElement.hpp"

ActingElement<Field, SEMILINEAR>& ActingElement<Field, SEMILINEAR>::mult(const ActingElement<Field, SEMILINEAR>& a,
        const ActingElement<
        Field, SEMILINEAR>& b, ActingElement<Field, SEMILINEAR>& res)
{
    Vec<Field> phi_copy = b.m_phi.permute(a.m_pi);
    phi_copy.applyFrobenius(a.m_frob_pow);
    Matrix<Field> A_copy(b.m_A);
    A_copy.applyFrobenius(a.m_frob_pow);

    res.m_phi = a.m_phi;
    res.m_phi *= phi_copy;
    res.m_pi = a.m_pi * b.m_pi;
    res.m_A = a.m_A * A_copy;
    res.m_frob_pow = (a.m_frob_pow + b.m_frob_pow) % Field::get_r();
    return res;
}

ActingElement<Field, SEMILINEAR> ActingElement<Field, SEMILINEAR>::operator*(const ActingElement& b) const
{
    ActingElement<Field, SEMILINEAR> res;
    return ActingElement<Field, SEMILINEAR>::mult(*this, b, res);
}

ActingElement<Field, SEMILINEAR>& ActingElement<Field, SEMILINEAR>::operator*=(const ActingElement& b)
{
    return ActingElement<Field, SEMILINEAR>::mult(*this, b, *this);
}

Matrix<Field> ActingElement<Field, SEMILINEAR>::operator*(const Matrix<Field>& Gamma) const
{
    Matrix<Field> res(Gamma);
    res.applyFrobenius(m_frob_pow);
    res = m_A * res;
    res.permuteCols(m_pi);
    unsigned int n = Gamma.getCols();
    for (unsigned int i = 0; i < n; ++i)
    {
        res.multiplyColumn(i, m_phi.get(i));
    }
    return res;
}

ActingElement<Field, SEMILINEAR>& ActingElement<Field, SEMILINEAR>::operator*=(const Vec<Field>& phi)
{
    Vec<Field> phi_copy = phi.permute(m_pi);
    phi_copy.applyFrobenius(m_frob_pow);
    m_phi *= phi_copy;
    return *this;
}

ActingElement<Field, SEMILINEAR> ActingElement<Field, SEMILINEAR>::inverse(void) const
{
    ActingElement<Field, SEMILINEAR> res;
    Permutation::inv(m_pi, res.m_pi);
    res.m_A = m_A.inverse();
    res.m_phi = m_phi.permute(res.m_pi);
    res.m_phi.multInverse();

    if (m_frob_pow != 0)
    {
        res.m_frob_pow = Field::get_r() - m_frob_pow;
        res.m_phi.applyFrobenius(res.m_frob_pow);
        res.m_A.applyFrobenius(res.m_frob_pow);
    }
    else
    {
        res.m_frob_pow = 0;
    }

    return res;
}

void ActingElement<Field, SEMILINEAR>::reset(unsigned int k, unsigned int n)
{
    ActingElement<Field, LINEAR>::reset(k, n);
    m_frob_pow = 0;
}

void ActingElement<Field, SEMILINEAR>::randomInit(unsigned int k, unsigned int n)
{
    ActingElement<Field, LINEAR>::randomInit(k, n);
    m_frob_pow = ((unsigned int) rand()) % Field::get_r();
}

std::ostream& operator<<(std::ostream& os, const ActingElement<Field, SEMILINEAR>& g)
{
    os << (const ActingElement<Field, LINEAR>&) g << std::endl;
    os << g.m_frob_pow << std::endl;
    return os;
}
