/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  actingElement.hpp
///  \brief Implements the acting group  \f$ GL_k(R) \times \left(  {R^*}^n  \wr S_n \right) \f$
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   2010-09-29
///
///  Copyright (C) 2010 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef ACTINGELEMENT_HPP_
#define ACTINGELEMENT_HPP_

#include <iostream>
#include "matrix.hpp"
#include "groups.h"
//#include "binvector.h"

template<typename RING, AlgTypeE ALGTYPE>
class ActingElement;

template<typename RING>
class ActingElement<RING, PERMUTATIONAL>
{
public:
    static My_Long_Int group_order(unsigned int n)
    {
        return help::factorial(n);
    }

    Matrix<RING> m_A;
    Permutation m_pi;

    void phi2multInverse(void){
        //do nothing, just a dummy
    }

    void m_A_applyOwnFrobenius(void)
    {
        //do nothing, just a dummy
    }

    static ActingElement& mult(const ActingElement& a, const ActingElement& b, ActingElement& res);
    ActingElement(const Permutation& pi = Permutation()) : m_pi(pi) {}
    ActingElement(const Matrix<RING>& A) : m_A(A) {}
    ActingElement(unsigned int k, unsigned int n,const Permutation& pi = Permutation()) :  m_pi(pi) {m_A.identity(k);}
    ActingElement(unsigned int k, const Vec<RING>& phi, const Permutation& pi = Permutation()) : m_pi(pi) {m_A.identity(k);}

    ActingElement operator*(const ActingElement& b) const;
    Matrix<RING> operator*(const Matrix<RING>& Gamma) const;
    ActingElement& operator*=(const ActingElement& b);

    ActingElement inverse(void) const;

    void reset(unsigned int k, unsigned int n);
    ActingElement& enlarge(unsigned int old_n, const std::vector<std::vector<unsigned int> >& new2oldIndices);

    void randomInit(unsigned int k, unsigned int n);

    template<typename R>
    friend std::ostream& operator<<(std::ostream& os, const ActingElement<R, PERMUTATIONAL>& g );
};



template<typename RING>
class ActingElement<RING, LINEAR>
{
public:
    static My_Long_Int group_order(unsigned int n)
    {
        My_Long_Int group_order = help::factorial(n);
        for (unsigned int i = 0; i < n; i++)
        {
            group_order *= RING::my_multiplicativeGroupSize();
        }
        return group_order;
    }

	Matrix<RING> m_A;
	Vec<RING> m_phi;
	Permutation m_pi;

    void phi2multInverse(void){
        m_phi.multInverse();
    }

    void m_A_applyOwnFrobenius(void)
    {
        //do nothing, just a dummy
    }

	static ActingElement& mult(const ActingElement& a, const ActingElement& b, ActingElement& res);
	ActingElement(const Permutation& pi = Permutation()) : m_pi(pi) {}
	ActingElement(const Matrix<RING>& A, const Vec<RING>& phi) : m_A(A), m_phi(phi) {}
	ActingElement(unsigned int k, const Vec<RING>& phi, const Permutation& pi = Permutation()) : m_phi(phi), m_pi(pi) {m_A.identity(k);}
	ActingElement(unsigned int k, unsigned int n, const Permutation& pi = Permutation()) : m_phi(n,1), m_pi(pi) {m_A.identity(k);}
	ActingElement operator*(const ActingElement& b) const;
	Matrix<RING> operator*(const Matrix<RING>& Gamma) const;
	ActingElement& operator*=(const ActingElement& b);
	ActingElement& operator*=(const Vec<RING>& phi);

	ActingElement inverse(void) const;
	void reset(unsigned int k, unsigned int n);

	ActingElement& enlarge(unsigned int old_n, const std::vector<std::vector<unsigned int> >& new2oldIndices); 

	void randomInit(unsigned int k, unsigned int n);

	unsigned int getFrob_pow(void) { return 0; } // just a dummy call

	template<typename R>
	friend std::ostream& operator<<(std::ostream& os, const ActingElement<R, LINEAR>& g );
};

template<>
class ActingElement<Field, SEMILINEAR> : public ActingElement<Field, LINEAR>
{
public:
    static My_Long_Int group_order(unsigned int n)
    {
        My_Long_Int group_order = ActingElement<Field, LINEAR>::group_order(n);
        group_order *= Field::get_r();
        return group_order;
    }

	unsigned int m_frob_pow;

    void m_A_applyOwnFrobenius(void)
    {
        m_A.applyFrobenius(m_frob_pow);
    }

	static ActingElement& mult(const ActingElement& a, const ActingElement& b, ActingElement& res);
	ActingElement(const Permutation& pi = Permutation()) : ActingElement<Field, LINEAR>(pi) {}
	ActingElement(const Matrix<Field>& A, const Vec<Field>& phi) : ActingElement<Field, LINEAR>(A, phi), m_frob_pow(0) {}
	ActingElement(unsigned int k, const Vec<Field>& phi, const Permutation& pi = Permutation()) :
		ActingElement<Field, LINEAR>(k, phi, pi), m_frob_pow(0)  {}
	ActingElement(unsigned int k, unsigned int n, const Permutation& pi = Permutation()) :
		ActingElement<Field, LINEAR>(k, n, pi), m_frob_pow(0)  {}
	ActingElement(unsigned int k, unsigned int n, unsigned int frob_pow ) :
		ActingElement<Field, LINEAR>(k, n), m_frob_pow(frob_pow)  {}
	ActingElement( const ActingElement<Field, LINEAR>& rhs) :
	        ActingElement<Field, LINEAR>(rhs), m_frob_pow(0)  {}

	ActingElement operator*(const ActingElement& b) const;
	Matrix<Field> operator*(const Matrix<Field>& Gamma) const;
	ActingElement& operator*=(const ActingElement& b);
	ActingElement& operator*=(const Vec<Field>& phi);

	ActingElement inverse(void) const;
	void reset(unsigned int k, unsigned int n);
//	ActingElement& enlarge(unsigned int old_n, const std::vector<std::vector<unsigned int> >& new2oldIndices);
	void randomInit(unsigned int k, unsigned int n);

	unsigned int getFrob_pow(void) { return m_frob_pow; }
	friend std::ostream& operator<<(std::ostream& os, const ActingElement<Field, SEMILINEAR>& g );
};

template<typename RING>
class ActingElement<RING, PLUSMINUSMONOMIAL> : public  ActingElement<RING, LINEAR>
{
    ActingElement(const ActingElement<RING, LINEAR>& rhs) : ActingElement<RING, LINEAR>(rhs) {}

public:
    static My_Long_Int group_order(unsigned int n)
    {
        My_Long_Int group_order = help::factorial(n);
        for (unsigned int i = 0; i < n; i++)
        {
            group_order *= 2;
        }
        return group_order;
    }


    ActingElement(const Permutation& pi = Permutation()) : ActingElement<RING, LINEAR>(pi) {}
    ActingElement(const Matrix<RING>& A, const Vec<RING>& phi) :  ActingElement<RING, LINEAR>(A, phi) {}
    ActingElement(unsigned int k, const Vec<RING>& phi, const Permutation& pi = Permutation()) : ActingElement<RING, LINEAR>(k, phi, pi) {}
    ActingElement(unsigned int k, unsigned int n, const Permutation& pi = Permutation()) : ActingElement<RING, LINEAR>(k,n,pi) {};

    ActingElement operator*(const ActingElement& b) const
    {
        ActingElement<RING, PLUSMINUSMONOMIAL> res(*this);
        res *= b;
        return res;
    }

    Matrix<Field> operator*(const Matrix<Field>& Gamma) const
    {
        return ActingElement<RING, LINEAR>::operator*(Gamma);
    }

    ActingElement& operator*=(const ActingElement& b)
    {
        ActingElement<RING, LINEAR>::operator*=(b);
        return *this;
    }
    ActingElement& operator*=(const Vec<RING>& phi)
    {
        ActingElement<RING, LINEAR>::operator*=(phi);
        return *this;
    }

    void randomInit(unsigned int k, unsigned int n);
    ActingElement inverse(void) const {
        ActingElement<RING, PLUSMINUSMONOMIAL> res(ActingElement<RING, LINEAR>::inverse());
        return res;
    }
 };


template<typename RING>
ActingElement<RING, PERMUTATIONAL>& ActingElement<RING, PERMUTATIONAL>::mult(
        const ActingElement<RING, PERMUTATIONAL>& a,
        const ActingElement<RING, PERMUTATIONAL>& b, ActingElement<RING, PERMUTATIONAL>& res)
{
    res.m_A = a.m_A * b.m_A;
    Permutation::mul(a.m_pi, b.m_pi, res.m_pi);
    return res;
}

template<typename RING>
ActingElement<RING, PERMUTATIONAL> ActingElement<RING, PERMUTATIONAL>::operator*(
        const ActingElement<RING, PERMUTATIONAL>& b) const
{
    ActingElement<RING, PERMUTATIONAL> res;
    mult(*this, b, res);
    return res;
}

template<typename RING>
Matrix<RING> ActingElement<RING, PERMUTATIONAL>::operator*(const Matrix<RING>& Gamma) const
{
    Matrix<RING> res(m_A * Gamma);
    res.permuteCols(m_pi);
    return res;
}

template<typename RING>
ActingElement<RING, PERMUTATIONAL>& ActingElement<RING, PERMUTATIONAL>::operator*=(
        const ActingElement<RING, PERMUTATIONAL>& b)
{
    this->m_A = this->m_A * b.m_A;
    Permutation::mul(this->m_pi, b.m_pi, this->m_pi);
    return *this;
}

template<typename RING>
ActingElement<RING, PERMUTATIONAL> ActingElement<RING, PERMUTATIONAL>::inverse(void) const
{
    ActingElement<RING, PERMUTATIONAL> res;
    Permutation::inv(m_pi, res.m_pi);
    res.m_A = m_A.inverse();
    return res;
}

template<typename RING>
void ActingElement<RING, PERMUTATIONAL>::reset(unsigned int k, unsigned int n)
{
    m_A.identity(k);
}

template<typename RING>
ActingElement<RING, PERMUTATIONAL>& ActingElement<RING, PERMUTATIONAL>::enlarge(
        unsigned int old_n, const std::vector<
        std::vector<unsigned int> >& new2oldIndices)
{
    std::vector<int> permlist(old_n, -1);
    unsigned int counter(0);
    int degree = m_pi.degree();
    std::vector<unsigned int> offset(new2oldIndices.size());

    for (unsigned int i = 0; i < new2oldIndices.size(); ++i)
    {
        offset[i] = counter;
        counter += new2oldIndices[i].size();
    }
    counter = 0;
    for (unsigned int i = 0; i < new2oldIndices.size(); ++i)
    {
        unsigned int myOffset = offset[m_pi.apply(i, degree)];
        for (unsigned int j = 0; j < new2oldIndices[i].size(); ++j)
        {
            permlist[counter] = myOffset + j;
            ++counter;
        }
    }
    m_pi.list(permlist);
    return *this;
}

template<typename RING>
void ActingElement<RING, PERMUTATIONAL>::randomInit(unsigned int k, unsigned int n)
{
    m_pi.random(n);
    m_A.identity(k);
}

template<typename RING>
std::ostream& operator<<(std::ostream& os, const ActingElement<RING, PERMUTATIONAL>& g)
{
    os << g.m_A << std::endl;
    os<<g.m_pi;
    os << std::endl;
    return os;
}


// it is allowed to pass a=res or b=res
template<typename RING>
ActingElement<RING, LINEAR>& ActingElement<RING, LINEAR>::mult(const ActingElement<RING, LINEAR>& a,
        const ActingElement<RING, LINEAR>& b, ActingElement<RING, LINEAR>& res)
{
	Vec<RING> phi_copy = b.m_phi.permute(a.m_pi);
	res.m_phi = a.m_phi;
	res.m_phi *= phi_copy;
	res.m_pi = a.m_pi * b.m_pi;
	res.m_A = a.m_A * b.m_A;
	return res;
}

template<typename RING>
ActingElement<RING, LINEAR> ActingElement<RING, LINEAR>::operator*(const ActingElement<RING, LINEAR>& b) const
{
	ActingElement<RING, LINEAR> res;
	return ActingElement<RING, LINEAR>::mult(*this, b, res);
}

template<typename RING>
ActingElement<RING, LINEAR>& ActingElement<RING, LINEAR>::operator*=(const ActingElement<RING, LINEAR>& b)
{
	return ActingElement<RING, LINEAR>::mult(*this, b, *this);
}

template<typename RING>
ActingElement<RING, LINEAR>& ActingElement<RING, LINEAR>::operator*=(const Vec<RING>& phi)
{
	Vec<RING> phi_copy = phi.permute(m_pi);
	m_phi *= phi_copy;
	return *this;
}

template<typename RING>
Matrix<RING> ActingElement<RING, LINEAR>::operator*(const Matrix<RING>& Gamma) const
{
	Matrix<RING> res;
	res = m_A * Gamma;
	res.permuteCols(m_pi);
	unsigned int n = Gamma.getCols();
	for(unsigned int i=0; i<n; ++i)
	{
		res.multiplyColumn(i, m_phi.get(i));
	}
	return res;
}

template<typename RING>
ActingElement<RING, LINEAR> ActingElement<RING, LINEAR>::inverse(void) const
{
	ActingElement<RING, LINEAR> res;
	Permutation::inv(m_pi, res.m_pi);
	res.m_A = m_A.inverse();
	res.m_phi = m_phi.permute(res.m_pi);
	res.m_phi.multInverse();
	return res;
}

template<typename RING>
void ActingElement<RING, LINEAR>::reset(unsigned int k, unsigned int n)
{
	m_phi = Vec<RING>(n,1);
	m_A.identity(k);
}

template<typename RING>
ActingElement<RING, LINEAR>& ActingElement<RING, LINEAR>::enlarge(unsigned int old_n,
        const std::vector<std::vector<unsigned int> >& new2oldIndices)
{
	Vec<RING> newphi(old_n, 1);
	std::vector<int> permlist(old_n, -1);
	unsigned int counter(0);
	int degree = m_pi.degree();
	std::vector<unsigned int> offset(new2oldIndices.size() );
	
	for(unsigned int i=0; i<new2oldIndices.size(); ++i)
	{
		offset[i] = counter;
		counter += new2oldIndices[i].size();
	}
	counter =0;
	for(unsigned int i=0; i<new2oldIndices.size(); ++i)
	{
		unsigned int myOffset = offset[m_pi.apply(i, degree)];
		Element<RING> mult = m_phi.get(i);
		for(unsigned int j=0; j<new2oldIndices[i].size(); ++j)
		{
			permlist[counter] =  myOffset + j;
			newphi.set(counter , mult) ;
			++counter;
		}
	}
//	print(std::cout, permlist, " ")<<std::endl;

	m_phi = newphi;
	m_pi.list(permlist);
	return *this;
}


template<typename RING>
void ActingElement<RING, LINEAR>::randomInit(unsigned int k, unsigned int n)
{
	m_pi.random(n);
	m_phi.randomInvertible(n);
	m_A.identity(k);
}

template<typename RING>
void ActingElement<RING, PLUSMINUSMONOMIAL>::randomInit(unsigned int k, unsigned int n)
{
    ActingElement<RING, LINEAR>::m_phi = Vec<RING>(n,1);
    typename Vec<RING>::iterator phiIter(ActingElement<RING, LINEAR>::m_phi.begin());
    typename Vec<RING>::iterator phiIterEnd(ActingElement<RING, LINEAR>::m_phi.end());

    Element<RING> minusOne( - Element<RING>(1) );
    for( ; phiIter<phiIterEnd  ; ++phiIter )
    {
        if(rand() & 1)
        {
            *phiIter = minusOne;
        }
    }
    ActingElement<RING, LINEAR>::m_pi.random(n);
    ActingElement<RING, LINEAR>::m_A.identity(k);
}

template<typename R>
std::ostream& operator<<(std::ostream& os, const ActingElement<R, LINEAR>& g )
{
	os<<g.m_A;
	os<<"["<<g.m_phi.get(0);
	for(unsigned int i=1; i<g.m_phi.size(); ++i)
	{
		os<<" "<<g.m_phi.get(i);
	}
	os<<"]"<<std::endl;
	os<<g.m_pi;
	return os;
}
#endif /*ACTINGELEMENT_HPP_*/
