/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// \file  groups.h
/// \brief Permutations, Groups and Actions
///
/// Here, everything needed for groups and group-actions is implemented.
///
/// Copyright (C) 2010 Ralf Gugisch
///	Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///	http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef GROUPS_H
#define GROUPS_H

#include<iostream>
#include<queue>
#include<map>
#include<algorithm>
#include<string>

////////////////////////////////////////////////////////////////////////////////
/// \class compare_indices
/// \brief sorting according to another vector
///
/// comparison function for eg. sorting indices acc. to another vector
/// (resulting in a permutation which would sort the vector)
///
///  -  vector<double> v; ...
///  -  vector<int> idx(v.size());   for(i=0; i<v-size(); ++i) idx[i]=i;
///  -
///  -  sort(idx.begin(), idx.end(), compare_indices<double>(v));
///
/// \sa Permutation::sorting_perm()
///
////////////////////////////////////////////////////////////////////////////////


template<typename T, typename CMP=std::less<T> >
class compare_indices:public std::binary_function<int, int, bool> {
	const T* x;
	CMP      cmp;
public:
	compare_indices(const T* x, CMP cmp)                      :x(x),cmp(cmp)   { }
	compare_indices(const T* x)                               :x(x),cmp()      { }
	bool operator()(int i1, int i2)                   { return cmp(x[i1],x[i2]); }
};





////////////////////////////////////////////////////////////////////////////////
/// \class Permutation
/// \brief A permutation is represented by the "second line".
///
/// Use like a vector<int>.
///
/// The vector representing the second line is only as big, as necessary,
/// i.e. the identity is represented by an empty vector. This way, I don't need
/// to know the degree of a group while scanning the generators.
///
/// \todo Maybe, it is worth, also to start the vector with the first nontrivial
/// mapped point, i.e. with the stabilizer-index?
///
/// - You m_can initialize permutations with:
///     - P.id()                        -> the identity
///     - P.random(n)                   -> a random permutation of degree n
///     - P.list(beg,end)               -> [a1,...,an]
///     - P.cycle(beg,end)              -> (a1,a2,...,an)
///     - P.stdcycle(n, a1)             -> (a1,a1+1,...,a1+n-1)
///     - P.transposition(a,b)          -> (a,b)
///     - P.sorting_perm(beg,end)       -> Get the sorting permutation, i.e.
///                                        p.apply(beg,end) sorts the container
///     - P.stable_sorting_perm(beg,end)-> dito, but using std::stable_sort()
///
///   You m_can even initialize giving the "second line" of an injective function
///   only, which will be completed to a permutation.
///
///     - P.fillup(beg,end)
///
///   There are static functions returning permutations, too:
///
///     - asPermutation(v)               -> reinterprete a vector<int> as Permutation
///     - listPermutation(n, a1,...,an)  -> [a1,...,an]
///     - cyclePermutation(n, a1,...,an) -> (a1,...,an)
///     - stdcycle(n, a1)                -> (a1,a1+1,...,a1+n-1)
///     - transposition(a,b)             -> (a,b)
///
/// - Apply P:
///     - P.apply(a)                   -> P[a], a is nonnegative integer
///             (also short: P[a])
///     - P.invapply(a)                -> P^-1[a]
///     - P.apply_copy(beg,end,ret)    -> [P[v[0]], P[v[1]], ...] (vector given with iterators)
///     - P.apply(beg,end)             -> [P[v[0]], P[v[1]], ...] (overwrite directly in v)
///     - P.permute_copy(v)            -> [v[P^-1[0]],v[P^-1[1]],...],
///     - P.permute(v)                 -> [v[P^-1[0]],v[P^-1[1]],...],
///             (v vector of length>= P.degree())
///
/// - Calculate with permutations
///    - g*h                           -> multiplication
///    - pow(g,h)                      -> g*h*g^-1
///    - inv(g)                        -> g^-1
///
///    There are also faster versions available of the form mul(a,b,ret).
///
////////////////////////////////////////////////////////////////////////////////

class Permutation { // Implementation must be _a_ vector<int>!
public:
	/// \relates Permutation
	/// A Permutation m_can be read and written in cycle- or list-notation.
	enum  PermNotation {cycle_notation, list_notation };
	static std::string LIST_SEP;

	///
	Permutation()                                                              { }

	///  Initialize with the list-representation of a permutation.
	Permutation(std::vector<int> const& v)                                 { list(v); }

	///
	Permutation& swap(Permutation& p)               { v.swap(p.v); return *this; }

	/// lexicographic comparison of the list representation
	int          compare(Permutation const& p) const;

	//
	// initializations
	//

	// Same as vector::reserve()
	void reserve(int i)                                          { v.reserve(i); }

	/// Initialize with the identity. (Synonym: clear())
	inline Permutation& id()                          { v.clear(); return *this; }

	/// Initialize with the identity. (Synonym: id())
	inline Permutation& clear()                       { v.clear(); return *this; }

	/// Initialize with the list-representation of a permutation.
	template<typename it1>
	Permutation& list(it1 b, it1 e)       { v.assign(b,e); trim(); return *this; }
	/// \overload
	Permutation& list(std::vector<int> const& v)    { return list(v.begin(),v.end()); }
	/// \overload
	//  Permutation& list(int len, int i0,int i1,...)    { return list(&i0,&i0+len); }

	/// Initialize with a transposition \f$(i,j)\f$.
	inline Permutation& transposition(int i, int j)
	{ init(std::max(i,j)+1); v[i]=j; v[j]=i; return *this; }

	/// Initialize with a cycle of the form \f$(i,i+1,...,i+len-1)\f$.
	inline Permutation& stdcycle(int len, int i0)
	{ int k=i0+len; init(k); v[--k]=i0; while(--k>=i0)v[k]=k+1; return *this; }

	/// Initialize with a general cycle.
	template<typename it1>
	Permutation& cycle(it1 b, it1 e);
	/// \overload
	Permutation& cycle(std::vector<int> const& vi){return cycle(vi.begin(),vi.end()); }
	/// \overload
	//  Permutation& cycle(int len, int i0,int i1,...)  { return cycle(&i0,&i0+len); }

	/// return the standard cycle (beg, beg+1, ..., end)
	Permutation& cycle(unsigned int beg, unsigned int end);
	
	/// Initialize with an injective vector and complete it to lexical smallest permutation
	template<typename it1>
	Permutation& fillup(it1 b, it1 e);
	/// \overload
	Permutation& fillup(std::vector<int> const& v){ return fillup(v.begin(),v.end()); }
	/// \overload
	//  Permutation& fillup(int len, int i0,int i1,...){ return fillup(&i0,&i0+len); }

	/// Initialize with a random permutation of n points.
	Permutation& random(int n);

	/// Get sorting permutation of a container
	template<typename IT1, typename CMP>
	Permutation& sorting_perm(IT1 beg, IT1 end, CMP cmp) {
		sorting_perm_inverse(beg, end, cmp);
		inv();
		return *this;
	}

	/// Get sorting permutation of a container, with default comparison
	template<typename IT1>
	Permutation& sorting_perm(IT1 beg, IT1 end){
		typedef typename std::iterator_traits<IT1>::value_type      _Value;
		sorting_perm_inverse(beg,end, std::less<_Value>());
		inv();
		return *this;
	}

	/// Get the inverse of sorting permutation of a container
	template<typename IT1, typename CMP>
	Permutation& sorting_perm_inverse(IT1 beg, IT1 end, CMP cmp) {
		typedef typename std::iterator_traits<IT1>::difference_type _Distance;
		typedef typename std::iterator_traits<IT1>::value_type      _Value;
		_Distance sz=end-beg;
		v.assign(sz,0);   for(int i=0; i<sz; ++i) v[i]=i;
		std::sort(v.begin(),v.end(),compare_indices<_Value,CMP>(&*beg,cmp));
		trim();
		return *this;
	}

	/// Get sorting permutation of a container, with default comparison
	template<typename IT1>
	Permutation& sorting_perm_inverse(IT1 beg, IT1 end){
		typedef typename std::iterator_traits<IT1>::value_type      _Value;
		return sorting_perm_inverse(beg,end, std::less<_Value>());
	}

	/// Get stable-sorting permutation of a container
	template<typename IT1, typename CMP>
	Permutation& stable_sorting_perm_inverse(IT1 beg, IT1 end, CMP cmp){
		typedef typename std::iterator_traits<IT1>::difference_type _Distance;
		typedef typename std::iterator_traits<IT1>::value_type      _Value;
		_Distance sz=end-beg;
		v.assign(sz,0);   for(int i=0; i<sz; ++i) v[i]=i;
		std::stable_sort(v.begin(),v.end(),compare_indices<_Value,CMP>(&*beg,cmp));
		trim();
		return *this;
	}

	/// Get stable-sorting permutation of a container, with default comparison
	template<typename IT1>
	Permutation& stable_sorting_perm_inverse(IT1 beg, IT1 end){
		typedef typename std::iterator_traits<IT1>::value_type      _Value;
		return stable_sorting_perm_inverse(beg,end, std::less<_Value>());
	}

	/// Get stable_sorting permutation of a container
	template<typename IT1, typename CMP>
	Permutation& stable_sorting_perm(IT1 beg, IT1 end, CMP cmp) {
		stable_sorting_perm_inverse(beg, end, cmp);
		inv();
		return *this;
	}

	/// Get stable_sorting permutation of a container, with default comparison
	template<typename IT1>
	Permutation& stable_sorting_perm(IT1 beg, IT1 end){
		typedef typename std::iterator_traits<IT1>::value_type      _Value;
		stable_sorting_perm_inverse(beg,end, std::less<_Value>());
		inv();
		return *this;
	}


	/// Test, if this is a valid (bijective) permuation.
	bool valid() const;

	//
	// Properties, Tests
	//

	/// The maximal \f$ i\in \mathbb{N} \f$, which is not fixed by p.
	int  degree() const                                       { return v.size(); }

	/// The stabilizer index of a permutation.
	int stab_idx(int i=0) const // i is a hint
	{
		int deg = degree();
		std::vector<int>::const_iterator v_it = v.begin();
		while(i<deg-1 && (*v_it==i || *v_it ==-1)) { ++i; v_it++;}
		return i;
	}

	/// The sign of a permutation: sign(id)=false
	/// You m_can calculate with signs with the ^-operator: sign(p*q) = sign(p)^sign(q)
	bool sign() const;

	/// Test, if p is trivial
	bool isId() const                                        { return v.empty(); }

	/// Test, if p is trivial (synonym for isId() )
	bool istrivial() const                                   { return v.empty(); }

	//
	// Calculation
	//

	/// mul(), div(), inv(), pow(), and standard operators *, /

	static Permutation& mul(Permutation const& A, Permutation const& B, Permutation &ret); 
	static Permutation& div(Permutation const& A, Permutation const& B, Permutation &ret); 
	static Permutation& pow(Permutation const& A, Permutation const& B, Permutation& ret); 
	static Permutation  pow(Permutation const& A, Permutation const& B) { Permutation ret; pow(A,B,ret); return ret;} 
	static Permutation& inv(Permutation const& A, Permutation &ret);             
	Permutation  operator* (Permutation const& B) const { Permutation ret; return mul(*this,B,ret); } 
	Permutation  operator/ (Permutation const& B) const { Permutation ret; return div(*this,B,ret); } 
	Permutation& operator*=(Permutation const& B)            { return mul(*this,B,*this); } 
	Permutation& operator/=(Permutation const& B)            { return div(*this,B,*this); } 
	Permutation& inv()                             { return inv(*this,*this);   } 


	/// ret = A * (t_min, t_max)
	static Permutation& mulTransposition(Permutation const& A, int t_min, int t_max, Permutation& ret);

	/// A <- (t_min, t_max) A
	static Permutation& mulTransposition(int t_min, int t_max, Permutation & A);

	//
	// Action on N
	//

	/// Apply p to a number \f$i\in \mathbb{N}\f$
	/// \sa apply()
	int operator[](int x) const                               { return apply_neu(x); }

	/// Apply p to a number \f$ i \in \mathbb{N} \f$.
	/// \sa operator[]
	int apply_neu   (int x) const            { return apply(x, degree()); }

	/// Apply p to a number \f$ i \in \mathbb{N} \f$.
	/// \sa operator[]
	int apply   (int x, int deg) const            {return (x>=deg ||v[x]<0)? x:v[x]; }

	/// Apply p to a number \f$ i \in \mathbb{N} \f$.
	/// \sa operator[]
	int apply2   (int x) const            { return v[x]; }

	/// Apply \f$p^{-1}\f$ to a number \f$i\in \mathbb{N}\f$
	int invapply_neu(int x) const
	{ int y=x, deg = degree(); 
	if(apply(y, deg)==x)
		return x;
	else //a cycle: x>= degree() or v[x]<0 can't happen, if however we get an infinite loop!
		while(v[y]!=x) y=v[y]; 
	return y; }

	/// Apply \f$p^{-1}\f$ to a number \f$i\in \mathbb{N}\f$
	int invapply(int x, int deg) const
	{ int y=x; 
	if(apply(y, deg)==x)
		return x;
	else //a cycle: x>= degree() or v[x]<0 can't happen, if however we get an infinite loop!
		while(v[y]!=x) y=v[y]; 
	return y; }

	/// Apply p to a container of int's.
	///
	/// \f$ p.apply([x_0,...,x_{n-1}]) = [p(x_0),\ldots,p(x_{n-1}]\f$
	///
	///  You may directly overwrite the input container (i.e. setting ret=b)
	template<typename it1, typename it2>
	void apply_copy(it1 b, it1 e, it2 ret) const
	{ for(;b!=e;++b,++ret) *ret=apply(*b); }

	/// Apply p to a container of int's.
	///
	/// \f$ p.apply([x_0,...,x_{n-1}]) = [p(x_0),\ldots,p(x_{n-1}]\f$
	///
	///  You may directly overwrite the input container (i.e. setting ret=b)
	template<typename it1>
	void apply(it1 b, it1 e) const                          { apply_copy(b,e,b); }

	/// Permute a container of arbitrary type.
	///
	/// \f$ p.permute([x_0,...,x_{n-1}]) = [x_{p^{-1}0},\ldots,x_{p^{-1}(n-1)}]\f$
	///
	/// output need to be a random access iterator; You may not overwrite the input
	/// iterator and you need to ensure, that p is really a permutation of the size
	/// of input container.
	template<typename it1, typename it2>
	inline void permute_copy(it1 b, it1 e, it2 ret) const
	{ for(int i=0;b!=e;++b,++i) ret[apply_neu(i)]=*b; }

	/// Permute a container of arbitrary type.
	///
	/// \f$ p.permute([x_0,...,x_{n-1}]) = [x_{p^{-1}0},\ldots,x_{p^{-1}(n-1)}]\f$
	///
	/// b must be random access iterator;
	template<typename it1>
	inline void permute(it1 b, it1 e) const {
		if(isId()){ return; }
		typename std::iterator_traits<it1>::difference_type n=e-b;
		std::vector<bool> touched(n,false);
		std::vector<bool>::iterator touched_i_it = touched.begin();
		int j, deg = degree();
		for(int i=0; i<n; ++i){ 
			if(!*touched_i_it) {
				*touched_i_it=true;
				for(j=apply(i, deg); j!=i; j=apply(j, deg)) {
					touched[j]=true;
					std::swap(b[i],b[j]);
				}
			}
			touched_i_it++;
		}	
		//  vector<T> tmp(b,e);  permute_copy(tmp.begin(),tmp.end(),b);
	}


	//
	// input/output
	//

	/// You m_can switch between cycle- and list- notation for writing
	static PermNotation print_style;

	/// print the permutation in cycle- or list- notation \sa print_style
	void print(std::ostream& out, unsigned int perm_offset,  PermNotation style=print_style) const
	{ (style==cycle_notation)? print_c(out, perm_offset):print_l(out, perm_offset); }

	/// print the permutation in list notation suitable for passing it to magma
	void magma_print(std::ostream& out, int n) const;

	/// scan a permutation from an istream
	bool scan(std::istream& in, unsigned int perm_offset)                  { return (scan_c(in, perm_offset) || scan_l(in, perm_offset)); }

	std::vector<int>::const_iterator getBegin(void) const
	{
		return v.begin();
	}

	//	  int  compare   (Permutation const& A, Permutation const& B) { return A.compare(B);    } 
	bool operator==(Permutation const& B) { return this->compare(B)!=0; } 
	bool operator< ( Permutation const& B) { return this->compare(B)< 0; } 
	bool operator> ( Permutation const& B) { return this->compare(B)> 0; } 
	bool operator<=( Permutation const& B) { return this->compare(B)<=0; } 
	bool operator>=( Permutation const& B) { return this->compare(B)>=0; }

	friend std::ostream&operator<<(std::ostream& out, Permutation const&B){ B.print(out,0); return out; }
//	const char* str(void) {                                            
//		std::ostringstream out; this->print(out); return out.str().c_str(); }

	friend std::istream&operator>>(std::istream& in, Permutation      &B)
	{ if(!B.scan(in,0)) in.setstate(std::ios::failbit); return in; }  
//	bool scan_str(char const* s) {                                 
//		std::istringstream in(s); return scan(in); }

private:
	friend Permutation listPermutation(int len, int,int,...);

	/// The data: v[i]==-1 has same effect as v[i]=i; v only as big as necessary
	std::vector<int> v;

	/// Initialize v to given length with [-1,...,-1] (same effect as [0,...,n-1])
	Permutation& init(int n)                     { v.assign(n,-1); return *this; }

	void print_c(std::ostream& out, unsigned int perm_offset) const;  ///< print in cycle notation
	void print_l(std::ostream& out, unsigned int perm_offset) const;  ///< print in list notation
	bool scan_c(std::istream& in, unsigned int perm_offset);          ///< scan  in cycle notation
	bool scan_l(std::istream& in, unsigned int perm_offset);          ///< scan  in list notation

	/// trim trivial entries from the back of v.
	void trim(){ int i=v.size(); while(--i>=0) if(v[i]!=i) break; v.resize(++i); }
};

/// Reinterprete a vector<int> as permutation (in list-notation). \relates Permutation
inline Permutation      & asPermutation(std::vector<int> &v)
{ return *reinterpret_cast<Permutation*>(&v); }

/// Reinterprete a vector<int> as permutation (in list-notation). \relates Permutation
inline Permutation const& asPermutation(const std::vector<int> &v)
{ return *reinterpret_cast<const Permutation*>(&v); }

/// Return a transposition (\a a,\a b). \sa Permutation::transposition() \relates Permutation
inline Permutation transposition(int a, int b)
{ Permutation P; return P.transposition(a,b); }

/// Return a cycle (\a a,\a a + 1,...) of lenght \a len. \sa Permutation::stdcycle() \relates Permutation
inline Permutation stdcycle(int len, int i0)
{ return Permutation().stdcycle(len,i0); }

/// Return a cycle \f$(* \a b, ...)\f$. \sa Permutation::cycle() \relates Permutation
template<typename it1>
inline Permutation cyclePermutation(it1 b, it1 e)             { return Permutation().cycle(b,e); }
/// Return a cycle \f$(v[0], ...)\f$. \sa Permutation::cycle() \relates Permutation
inline Permutation cyclePermutation(std::vector<int> const& vi)
{ Permutation P; P.cycle(vi); return P; }

/// Return a cycle \f$(i0, i1, ...)\f$. \sa Permutation::cycle() \relates Permutation
Permutation cyclePermutation(int len, int i0,int i1,...);

/// Return a Permutation given as list \relates Permutation
Permutation listPermutation(int len, int i0,int i1,...);

/// Fills given list up to a complete Permutation \relates Permutation
Permutation fillupPermutation(int len, int i0,int i1,...);


template<typename it1>
Permutation& Permutation::cycle(it1 b, it1 e) {
	int k;
	init(*std::max_element(b,e)+1);
	v[*--e]=*b; k=*e; while(e--!=b) { v[*e]=k; k=*e; }
	return *this;
}

template<typename it1>
Permutation& Permutation::fillup(it1 b, it1 e) {
	v.assign(b,e);
	// fill up gaps in the vector (if any):
	std::vector<int> used(v);
	std::sort(used.begin(),used.end());
	int gaps=used.back()-used.size()+1;
	for(int i=0,x=0; gaps>0; --gaps) {
		while(x==used[i]) { ++x; ++i; }
		v.push_back(x++);
	}
	trim();
	return *this;
}
#endif // GROUPS_H
