/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  ringelement.hpp
///  \brief Implements an element in a chain ring.
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   2010-09-29
///
///  Copyright (C) 2010 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


#ifndef RINGELEMENT_HPP_
#define RINGELEMENT_HPP_

#define  INDEX unsigned int

#include <iostream>
#include <vector>
class Z4;
class F2_F2;
class Field;

template<typename RING>
class Element
{
private:
	INDEX e;
public:
	inline Element(INDEX _e=0);
	inline unsigned int height(void) const;
	inline Element multInverse(void) const;
	inline bool isOne(void) const;
	inline bool isZero(void) const;
	inline bool isUnit(void) const;
	inline bool isReducable(void) const;
	inline unsigned int getIndex(void) const;
	
	template<typename R>
	friend std::ostream& operator<<(std::ostream& os, const Element<R>& a);
	
	std::ostream& printBasisRep(std::ostream& os,  char basis_symbol ) const;

	inline int compare(const Element& b) const;
	
	inline Element& operator=(const Element& b);
	
	inline bool operator==(const Element& b) const;
	inline bool operator<(const Element& b) const;
	inline bool operator<=(const Element& b) const;
	inline bool operator>(const Element& b) const;
	inline bool operator>=(const Element& b) const;
	inline bool operator!=(const Element& b) const;
	
	inline Element operator+(const Element& b) const;
	inline Element operator-(const Element& b) const;
	inline Element operator-(void) const;
	inline Element operator*(const Element& b) const;
	inline Element operator/(const Element& b) const;
	
	inline Element& operator+=(const Element& b);
	inline Element& operator-=(const Element& b);
	inline Element& operator*=(const Element& b);
	inline Element& operator/=(const Element& b);

	// only for finite fields
	Element& applyFrobenius(unsigned int pow);
	unsigned int minimizeByFrobenius(unsigned int& pow);

	const std::vector<unsigned int>& getBasisRepresentation(void) const;
};

template<typename RING>
Element<RING>::Element(INDEX _e)
{	
	e = _e;
}

template<typename RING>
inline unsigned int Element<RING>::height(void) const
{	
	return RING::height(e);
}

template<typename RING>
Element<RING> Element<RING>::multInverse(void) const
{
	return RING::multInverse(e);
}

template<typename RING>
bool Element<RING>::isOne(void) const
{
	return (e==1);
}

template<typename RING>
bool Element<RING>::isZero(void) const
{
	return (e==0);
}

template<typename RING>
bool Element<RING>::isUnit(void) const
{
	return RING::isUnit(*this);
}

template<typename RING>
bool Element<RING>::isReducable(void) const
{
	return RING::isReducable(*this);
}

template<typename RING>
unsigned int Element<RING>::getIndex(void) const
{
	return e;
}

template<typename RING>
int Element<RING>::compare(const Element& b) const
{	
	return RING::compare(*this, b);
}

template<typename RING>
Element<RING> & Element<RING>::operator=(const Element<RING>& b)
{
	e = b.e;
	return *this;
}

template<typename RING>
bool Element<RING>::operator==(const Element<RING>& b) const
{
	return e == b.e;
}

template<typename RING>
bool Element<RING>::operator<(const Element<RING>& b) const
{
	return compare(b) < 0;
}

template<typename RING>
bool Element<RING>::operator<=(const Element<RING>& b) const
{
	return compare(b) <= 0;
}

template<typename RING>
bool Element<RING>::operator>(const Element<RING>& b) const
{
	return compare(b) > 0;
}

template<typename RING>
bool Element<RING>::operator>=(const Element<RING>& b) const
{
	return compare(b) >= 0;
}

template<typename RING>
bool Element<RING>::operator!=(const Element<RING>& b) const
{
	return e  != b.e;
}

template<typename RING>
Element<RING> Element<RING>::operator+(const Element<RING>& b) const
{	
	return Element<RING>(RING::add(*this, b));
}

template<typename RING>
Element<RING> Element<RING>::operator-(const Element<RING>& b) const
{	
	return Element<RING>(RING::minus(*this, b));
}

template<typename RING>
Element<RING> Element<RING>::operator-(void) const
{	
	return Element<RING>(RING::addInverse(*this));
}

template<typename RING>
Element<RING> Element<RING>::operator*(const Element<RING>& b) const
{	
	return RING::mult(*this, b);
}

template<typename RING>
Element<RING> Element<RING>::operator/(const Element<RING>& b) const
{	
	return Element<RING>(RING::div(*this, b));
}

template<typename RING>
Element<RING>& Element<RING>::operator+=(const Element<RING>& b)
{
	e = RING::add(*this, b).getIndex();
	return *this;
}

template<typename RING>
Element<RING>& Element<RING>::operator-=(const Element<RING>& b)
{
	e = RING::minus(*this, b);
	return *this;
}

template<typename RING>
Element<RING>& Element<RING>::operator*=(const Element<RING>& b)
{
	e = RING::mult(*this, b).getIndex();
	return *this;
}

template<>
Element<Z4>& Element<Z4>::operator*=(const Element<Z4>& b);

template<>
Element<F2_F2>& Element<F2_F2>::operator*=(const Element<F2_F2>& b);


//template<>
//Element<F2_F2>& Element<F2_F2>::operator*=(const Element<F2_F2>& b)
//{
//	e = RING::mult(e, b.e);
//	return *this;
//}

template<typename RING>
Element<RING>& Element<RING>::operator/=(const Element<RING>& b)
{
	e = RING::div(*this, b);
	return *this;
}

template<>
Element<Field>& Element<Field>::applyFrobenius(unsigned int pow);

template<>
unsigned int Element<Field>::minimizeByFrobenius(unsigned int& pow);

template<typename RING>
std::ostream& operator<<(std::ostream& os, const Element<RING>& a)
{
	return RING::print(os, a);
}

template<typename RING>
std::ostream& Element<RING>::printBasisRep(std::ostream& os,  char basis_symbol ) const
{
    return RING::print(os, *this);
}

template<>
const std::vector<unsigned int>& Element<Field>::getBasisRepresentation(void) const;

template<>
std::ostream& Element<Field>::printBasisRep(std::ostream& os,  char basis_symbol ) const;


#endif /*RINGELEMENT_HPP_*/
