/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  ringelement.hpp
///  \brief Implements a vector over a chain ring.
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   2010-09-29
///
///  Copyright (C) 2010 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


#ifndef VECTOR_HPP_
#define VECTOR_HPP_

#include <vector>
#include "defs.h"
#include "ringelement.hpp"
#include "helpfunction.hpp"
#include "groups.h"

class Field;
class BinaryField;
class F2_F2;
class GF2r;


template<typename RING>
class Matrix;

template<unsigned int chainlength>
class SymWeight;

template<typename RING>
class Vec: public std::vector<Element<RING> >
{
public:
    Vec(int i = 0, Element<RING> el = 0);
    Vec(const std::vector<Element<RING> >& v) :
        std::vector<Element<RING> >(v)
        {
        }
    // returns <height, entry_index>
    inline std::pair<unsigned int, unsigned int> height(unsigned int startpos = 0) const;

    inline void set(unsigned int i, const Element<RING>& a)
    {
        this->operator[](i) = a;
    }

    inline const Element<RING>& get(unsigned int i) const
                    {
        return this->operator[](i);
                    }

    inline void swap(unsigned int ent1, unsigned int ent2)
    {
        std::swap(this->operator[](ent1), this->operator[](ent2));
    }

    using std::vector<Element<RING> >::push_back;
    using std::vector<Element<RING> >::back;
    using std::vector<Element<RING> >::begin;
    using std::vector<Element<RING> >::end;
    using std::vector<Element<RING> >::rbegin;
    using std::vector<Element<RING> >::rend;
    using std::vector<Element<RING> >::size;
    using std::vector<Element<RING> >::iterator;
    using std::vector<Element<RING> >::const_iterator;
    using std::vector<Element<RING> >::const_reverse_iterator;

    //	unsigned int getLeeWeight(void) const;
    //	static SymWeight getSymWeight(const Vec& v) {return v.getSymWeight();}
    inline std::pair<unsigned int, Element<RING> > getFirstUnit(void) const;

    inline Element<RING> normalize(bool onlyByMinusOne);
    inline bool isNormalized(unsigned int k1) const;

    inline void randomInvertible(unsigned int len);

    inline Element<RING> scalar(const Vec& b) const;
    inline void multEntry(unsigned int pos, const Element<RING>& el);

    inline Vec& operator+=(const Vec& b);
    inline Vec& operator-=(const Vec& b);
    inline Vec& operator*=(const Vec& b);
    inline Vec& operator*=(const Element<RING>& x);

    inline Vec operator+(const Vec& b) const;
    inline Vec operator-(const Vec& b) const;
    inline Vec operator*(const Vec& b) const;
    inline Vec operator*(const Element<RING>& x) const;

    inline int compare(const Vec& b) const;
    inline bool operator<(const Vec& b) const;
    inline bool operator<=(const Vec& b) const;
    inline bool operator>(const Vec& b) const;
    inline bool operator>=(const Vec& b) const;
    inline bool operator==(const Vec& b) const;

    inline Vec<RING>& permuteReplace(const Permutation& pi);
    inline Vec<RING> permute(const Permutation& pi) const;
    inline Vec<RING>& multInverse(void);

    inline unsigned int fill(unsigned int rank, bool projective);

    template<typename R>
    friend std::ostream& operator<<(std::ostream& os, const Vec<R>& v);

    std::ostream& print(std::ostream& os, const std::string& functionName) const;

    unsigned int getRank(unsigned int len) const;

    /// only for a finite field
    inline void applyFrobenius(unsigned int frob_pow);

    static Element<RING> innerProduct(const Vec<RING>& a, const Vec<RING>& b);
};

template<typename RING>
class DenseVec: public Vec<RING>
{
public:
    DenseVec<RING>(int i = 0, Element<RING> el = 0) : Vec<RING>(i, el)
    {}

    DenseVec<RING>(const Vec<RING>& v) : Vec<RING> (v)
                    {    }

    DenseVec<Field> operator*(const Element<Field>& x) const;
    SymWeight<1u> getWeight(void) const;
    unsigned int grayImageWeight(void) const;
};

template<>
class DenseVec<BinaryField>
{
private:
    static char bits_in_16bits[0x1u<<16];
    static int iterated_bitcount(uchar n);


public:
    unsigned int m_len;
    std::vector<uchar> m_data;

    static const unsigned int BITPOSMASK = 63; // get the last bits
    static void compute_bits_in_16bits(void);
    static int bit_count(uchar n);

    DenseVec<BinaryField>(int i = 0);
    DenseVec<BinaryField>(const std::vector<Element<BinaryField> >& v);

    template<typename RING>
    DenseVec<BinaryField>(const Matrix<RING>& m);

    bool operator[](unsigned int i) const
    {
        return get(i);
    }
    bool get(unsigned int i) const
    {
        return (m_data[i>>6] >> (i & BITPOSMASK)) & 1;
    }

    bool getBits(unsigned int datapos, uchar bitmask) const
    {
        return m_data[datapos] & bitmask;
    }

    void setBit(unsigned int datapos, unsigned int bitpos)
    {
        m_data[datapos] |= ((uchar) 1) << bitpos;
    }
    void setBit(unsigned int i)
    {
        setBit( i >> 6,  i & BITPOSMASK);
    }

    SymWeight<1u> getWeight(const DenseVec& mask) const;
    SymWeight<1u> getWeight(void) const;
    unsigned int grayImageWeight(void) const;
    DenseVec& operator+=(const DenseVec& b);
    DenseVec& operator|=(const DenseVec& b);
    DenseVec& operator&=(const DenseVec& b);
    DenseVec operator~(void) const;
    DenseVec operator<<(unsigned int);
    //    inline Codeword& operator*=(const Element<BinaryField>& x);

    void clear(void);

    friend std::ostream& operator<<(std::ostream& os, const DenseVec<BinaryField>& c);
    //    void getHeightSeq(std::vector<unsigned int>::iterator iterRes, std::vector<unsigned int>::iterator iterResEnd) const;
    std::vector<uchar>::const_iterator begin(void)
    {
        return m_data.begin();
    }
    std::vector<uchar>::const_iterator end(void)
    {
        return m_data.end();
    }

    inline int compare(const DenseVec<BinaryField>& b) const
    {
        if(m_len != b.m_len)
        {
            return m_len - b.m_len;
        }
        std::vector<uchar>::const_iterator iter(m_data.begin());
        std::vector<uchar>::const_iterator iterEnd(m_data.end());
        std::vector<uchar>::const_iterator iterB(b.m_data.begin());
        for (; iter < iterEnd; ++iter, ++iterB)
        {
            if(*iter != *iterB)
            {
                if( *iter < *iterB)
                {
                    return -1;
                }
                return 1;
            }
        }
        return 0;
    }

    bool operator<(const DenseVec<BinaryField>& b) const
    {
        return compare(b)<0;
    }

    bool operator==(const DenseVec& b) const
                    {
        return compare(b)!=0;
                    }

    bool scalar(const DenseVec<BinaryField>& v) const
    {
        bool res(false);
        unsigned int words(m_data.size());
        for (unsigned int i = 0; i <words ; ++i)
        {
            res ^= bit_count(m_data[i] & v.m_data[i]) & 1;
        }
        return res;
    }

    void write(std::ofstream& os) const
    {
        std::vector<uchar>::const_iterator iter(m_data.begin());
        std::vector<uchar>::const_iterator iterEnd(m_data.end());
        for (; iter < iterEnd; ++iter)
        {
            os.write((char*) &*iter, sizeof(uchar));
        }
    }

    bool read(std::ifstream& in)
    {
        std::vector<uchar>::const_iterator iter(m_data.begin());
        std::vector<uchar>::const_iterator iterEnd(m_data.end());
        for (; iter < iterEnd && in.good(); ++iter)
        {
            in.read((char*) &*iter, sizeof(uchar));
        }
        return in.good();
    }

    class Iterator
    {
    private:
        const DenseVec<BinaryField>* m_vec;
        unsigned int m_data_pos, m_bit_pos;

        bool check(void)
        {
            return getIndex() < m_vec->m_len;
        }

    public:
        Iterator(const DenseVec<BinaryField>* const v) : m_vec(v), m_data_pos(0), m_bit_pos(0) {}
        bool operator+=(unsigned int i)
                {
            m_bit_pos += i;
            if(check())
            {
                while(m_bit_pos >  BITPOSMASK)
                {
                    m_bit_pos -= (BITPOSMASK+1);
                    ++m_data_pos;
                }
                return true;
            }
            else
            {
                return false;
            }
                }

        bool operator++(void)
                {
            if( m_bit_pos == BITPOSMASK)
            {
                m_bit_pos = 0;
                ++m_data_pos;
            }
            else
            {
                ++m_bit_pos;
            }
            return check();
                }

        bool operator*(void) const
        {
            return (m_vec->m_data[m_data_pos]>>m_bit_pos) & 1;
        }

        unsigned int getIndex(void) const
        {
            return (m_data_pos<<6)+m_bit_pos;
        }

        unsigned int get_data_pos(void) const
        {
            return m_data_pos;
        }
        unsigned int get_bit_pos(void) const
        {
            return m_bit_pos;
        }
    };

    DenseVec<BinaryField>::Iterator getIterator(void) const
    {
        return DenseVec<BinaryField>::Iterator(this);
    }

    //    static void normalize( std::vector<Codeword<BinaryField> >& words, unsigned int checksize, const std::vector<unsigned int> pos)
    //    {
    //        return; // they are normalized (just a dummy)
    //    }
};

template<>
class DenseVec<GF2r>
{
private:
    unsigned int m_len;
    std::vector<DenseVec<BinaryField> > m_data;
public:
    DenseVec(int i = 0);
    DenseVec(const std::vector<Element<GF2r> >& v);
};

/// \brief Vectors over the integers modulo 4.
///
/// 3 bits represent one entry. The third bit is used to fasten the arithmetical operations.
template<>
class DenseVec<Z4>
{
private:
    unsigned int m_len;
    std::vector<uchar> m_data;
    static const uchar CLEARMASK = 0333333333333333333333ll; //octal
    static const uchar ONEBITMASK = 0111111111111111111111ll; // octal

public:
    std::vector<Element<Z4> > getVector(void) const;

    static const unsigned int ENTRIES_PER_WORD = 21;
    DenseVec<Z4>(int i = 0);
    DenseVec<Z4>(const std::vector<Element<Z4> >& v);

    inline Element<Z4> operator[](unsigned int i) const;
    inline void setBits(unsigned int i);
    SymWeight<2u> getWeight(const DenseVec& mask) const;
    std::vector<unsigned int> getCompleteWeight(const DenseVec& mask) const;
    SymWeight<2u> getWeight(void) const;
    unsigned int grayImageWeight(void) const;
    inline DenseVec& operator+=(const DenseVec& b);
    inline DenseVec& operator*=(const Element<Z4>& x);

    friend std::ostream& operator<<(std::ostream& os, const DenseVec<Z4>& c);
    void getHeightSeq(std::vector<unsigned int>::iterator iterRes, std::vector<unsigned int>::iterator iterResEnd) const;
    std::vector<uchar>::const_iterator begin(void)
    {
        return m_data.begin();
    }
    std::vector<uchar>::const_iterator end(void)
    {
        return m_data.end();
    }
    uchar getData(unsigned int pos) const
    {
        return m_data[pos];
    }

    bool operator==(const DenseVec& b) const
                    {
        std::vector<uchar>::const_iterator iter(m_data.begin());
        std::vector<uchar>::const_iterator iterEnd(m_data.end());
        std::vector<uchar>::const_iterator iterB(b.m_data.begin());
        for (; iter < iterEnd; ++iter, ++iterB)
        {
            if(*iter != *iterB)
            {
                return false;
            }
        }
        return true;
                    }
};

/// \brief Vectors over the chain ring F2_F2
///
/// 2 bits represent one entry. There are 32 entries per entry in m_data
template<>
class DenseVec<F2_F2>
{
private:
    unsigned int m_len;
    std::vector<uchar> m_data;
    static const uchar ONEBITMASK = 0525252525252525252525ll; // octal

public:
    static const unsigned int BITPOSMASK = 31; // used for modulo 32
    std::vector<Element<F2_F2> > getVector(void) const;
    DenseVec<F2_F2>(int i = 0);
    DenseVec<F2_F2>(const std::vector<Element<F2_F2> >& v);

    inline Element<F2_F2> operator[](unsigned int i) const;
    inline void setBits(unsigned int i);
    SymWeight<2u> getWeight(const DenseVec& mask) const;
    std::vector<unsigned int> getCompleteWeight(const DenseVec& mask) const;
    SymWeight<2u> getWeight(void) const;
    unsigned int grayImageWeight(void) const;
    inline DenseVec& operator+=(const DenseVec& b);
    inline DenseVec& operator*=(const Element<F2_F2>& x);

    friend std::ostream& operator<<(std::ostream& os, const DenseVec<F2_F2>& c);
    void getHeightSeq(std::vector<unsigned int>::iterator iterRes, std::vector<unsigned int>::iterator iterResEnd) const;
    std::vector<uchar>::const_iterator begin(void)
    {
        return m_data.begin();
    }
    std::vector<uchar>::const_iterator end(void)
    {
        return m_data.end();
    }
    uchar getData(unsigned int pos) const
    {
        return m_data[pos];
    }

    bool operator==(const DenseVec& b) const
                    {
        std::vector<uchar>::const_iterator iter(m_data.begin());
        std::vector<uchar>::const_iterator iterEnd(m_data.end());
        std::vector<uchar>::const_iterator iterB(b.m_data.begin());
        for (; iter < iterEnd; ++iter, ++iterB)
        {
            if(*iter != *iterB)
            {
                return false;
            }
        }
        return true;
                    }
};

template<typename RING>
std::ostream& operator<<(std::ostream& os, const Vec<RING>& v)
{
    return help::print(os, v, " ");
}

template<typename RING>
std::ostream& Vec<RING>::print(std::ostream& os, const std::string& functionName) const
{
    if(this->size()==0)
    {
        os<<"()";
        return os;
    }

    typename Vec<RING>::const_iterator iter(this->begin());
    typename Vec<RING>::const_iterator iterEnd(this->end());
    os<<functionName<<"("<<*iter<<")";
    ++iter;
    for (; iter < iterEnd; ++iter)
    {
        os<<", "<<functionName<<"("<<*iter<<")";
    }
    return os;
}

template<typename RING>
Vec<RING>::Vec(int i, Element<RING> el) :
std::vector<Element<RING> >(i, el)
{
}

template<typename RING>
int Vec<RING>::compare(const Vec<RING>& b) const
{
    //	both vectors are assumed to have the same size
    if (this->size() < b.size())
    {
        return -1;
    }
    if (this->size() > b.size())
    {
        return 1;
    }
    typename Vec<RING>::const_reverse_iterator iter(this->rbegin());
    typename Vec<RING>::const_reverse_iterator iterEnd(this->rend());
    typename Vec<RING>::const_reverse_iterator iterB(b.rbegin());

    for (; iter < iterEnd; ++iter, ++iterB)
    {
        if (*iter < *iterB)
        {
            return -1;
        }
        if (*iter > *iterB)
        {
            return 1;
        }
    }
    return 0;
}

template<typename RING>
std::pair<unsigned int, unsigned int> Vec<RING>::height(unsigned int startpos) const
{
    typename Vec<RING>::const_iterator iter(this->begin() + startpos);
    typename Vec<RING>::const_iterator iterEnd(this->end());
    std::pair<unsigned int, unsigned int> res(RING::chainlength(), startpos);

    while (iter < iterEnd)
    {
        if (iter->height() < res.first)
        {
            res.first = iter->height();
            res.second = distance(this->begin(), iter);
            if (res.first == 0)
            {
                return res;
            }
        }
        ++iter;
    }
    return res;
}

template<typename RING>
std::pair<unsigned int,  Element<RING> > Vec<RING>::getFirstUnit(void) const
{
    typename Vec<RING>::const_iterator iter(this->begin());
    typename Vec<RING>::const_iterator iterEnd(this->end());

    while (iter < iterEnd)
    {
        if (iter->isUnit())
        {
            return std::pair<unsigned int,  Element<RING> >( distance(this->begin(), iter), *iter);
        }
        ++iter;
    }
    return std::pair<unsigned int,  Element<RING> >(this->size(),0);
}

template<typename RING>
Element<RING> Vec<RING>::normalize(bool onlyByMinusOne)
{
    typename Vec<RING>::reverse_iterator iter(this->rbegin());
    typename Vec<RING>::reverse_iterator iterEnd(this->rend());

    Element<RING> mult(0);

    while (iter < iterEnd)
    {
        if (iter->isUnit())
        {
            if (mult.isZero())
            {
                if(onlyByMinusOne)
                {
                    if(-(*iter)<*iter)
                    {
                        *iter = -(*iter);
                        mult = - Element<RING>(1);
                    }
                    else
                    {
                        mult = Element<RING>(1);
                    }
                }
                else
                {
                    mult = iter->multInverse();
                    *iter = Element<RING> (1);
                }
            }
            else
            {
                *iter *= mult;
            }
        }
        ++iter;
    }
    if (mult == 0)
    {
        return Element<RING> (1);
    }
    return mult;
}

template<typename RING>
bool Vec<RING>::isNormalized(unsigned int k1) const
{
    typename Vec<RING>::const_reverse_iterator iter(this->rbegin() + k1);
    typename Vec<RING>::const_reverse_iterator iterEnd(this->rend());

    while (iter < iterEnd)
    {
        if (iter->isUnit())
        {
            if (iter->isOne())
            {
                return true;
            }
            return false;
        }

        ++iter;
    }
    return true;
}

template<typename RING>
void Vec<RING>::randomInvertible(unsigned int len)
{
    this->resize(len);
    typename std::vector<Element<RING> >::iterator iter(this->begin());
    typename std::vector<Element<RING> >::iterator iterEnd(this->end());
    for (; iter < iterEnd; ++iter)
    {
        *iter = RING::randomInvertible();
    }
}

template<typename RING>
Element<RING> Vec<RING>::scalar(const Vec& b) const
{
    Element<RING> res(0);
    typename Vec<RING>::const_iterator iter(this->begin());
    typename Vec<RING>::const_iterator iterEnd(this->end());
    typename Vec<RING>::const_iterator bIter(b.begin());

    while (iter < iterEnd)
    {
        res += (*iter) * (*bIter);
        ++iter;
        ++bIter;
    }
    return res;
}

template<typename RING>
void Vec<RING>::multEntry(unsigned int pos, const Element<RING>& el)
{
    this->operator[](pos) *= el;
}

template<typename RING>
Vec<RING>& Vec<RING>::operator+=(const Vec<RING>& b)
{
    typename Vec<RING>::iterator iter(this->begin());
    typename Vec<RING>::const_iterator iterb(b.begin());
    typename Vec<RING>::const_iterator iterEnd(this->end());

    for (; iter < iterEnd; ++iter, ++iterb)
    {
        *iter += *iterb;
    }
    return *this;
}

template<typename RING>
Vec<RING> Vec<RING>::operator+(const Vec<RING>& b) const
{
    Vec<RING> res(*this);
    res += b;
    return res;
}

template<typename RING>
Vec<RING>& Vec<RING>::operator-=(const Vec<RING>& b)
{
    typename Vec<RING>::iterator iter(this->begin());
    typename Vec<RING>::const_iterator iterb(b.begin());
    typename Vec<RING>::const_iterator iterEnd(this->end());

    for (; iter < iterEnd; ++iter, ++iterb)
    {
        *iter -= *iterb;
    }
    return *this;
}

template<typename RING>
Vec<RING> Vec<RING>::operator-(const Vec<RING>& b) const
{
    Vec<RING> res(*this);
    res -= b;
    return res;
}

template<typename RING>
Vec<RING>& Vec<RING>::operator*=(const Vec<RING>& b)
{
    typename Vec<RING>::iterator iter(this->begin());
    typename Vec<RING>::const_iterator iterb(b.begin());
    typename Vec<RING>::const_iterator iterEnd(this->end());

    for (; iter < iterEnd; ++iter, ++iterb)
    {
        *iter *= *iterb;
    }
    return *this;
}

template<typename RING>
Vec<RING> Vec<RING>::operator*(const Vec<RING>& b) const
{
    Vec<RING> res(*this);
    res *= b;
    return res;
}

template<typename RING>
Vec<RING>& Vec<RING>::operator*=(const Element<RING>& x)
{
    typename Vec<RING>::iterator iter(this->begin());
    typename Vec<RING>::const_iterator iterEnd(this->end());

    for (; iter < iterEnd; ++iter)
    {
        *iter *= x;
    }
    return *this;
}

template<typename RING>
Vec<RING> Vec<RING>::operator*(const Element<RING>& x) const
{
    Vec<RING> res(*this);
    res *= x;
    return res;
}

template<typename RING>
Element<RING> Vec<RING>::innerProduct(const Vec<RING>& a, const Vec<RING>& b)
{
    Element<RING> res(0);
    typename Vec<RING>::const_iterator aIter(a.begin());
    typename Vec<RING>::const_iterator bIter(b.begin());
    typename Vec<RING>::const_iterator aIterEnd(a.end());

    for (; aIter < aIterEnd; ++aIter, ++bIter)
    {
        res += *aIter * *bIter;
    }
    return res;
}


template<typename RING>
bool Vec<RING>::operator<(const Vec<RING>& b) const
{
    return compare(b) == -1;
}
template<typename RING>
bool Vec<RING>::operator<=(const Vec<RING>& b) const
{
    return compare(b) <= 0;
}
template<typename RING>
bool Vec<RING>::operator>(const Vec<RING>& b) const
{
    return compare(b) == 1;
}
template<typename RING>
bool Vec<RING>::operator>=(const Vec<RING>& b) const
{
    return compare(b) >= 0;
}

template<typename RING>
bool Vec<RING>::operator==(const Vec<RING>& b) const
{
    return compare(b) == 0;
}

template<typename RING>
Vec<RING>& Vec<RING>::permuteReplace(const Permutation& pi)
{
    pi.permute(this->begin(), this->end());
    return *this;
}

template<typename RING>
Vec<RING> Vec<RING>::permute(const Permutation& pi) const
{
    Vec<RING> res(*this);
    pi.permute(res.begin(), res.end());
    return res;
}

template<typename RING>
inline Vec<RING>& Vec<RING>::multInverse(void)
{
    typename Vec<RING>::iterator iter(this->begin());
    typename Vec<RING>::const_iterator iterEnd(this->end());

    for (; iter < iterEnd; ++iter)
    {
        *iter = iter->multInverse();
    }
    return *this;
}

template<typename RING>
inline unsigned int Vec<RING>::fill(unsigned int rank, bool projective)
{
    typename Vec<RING>::iterator iter(this->begin());
    typename Vec<RING>::iterator iterEnd;
    if(projective)
    {
        iterEnd = (iter + (this->size() -1) );
    }
    else
    {
        iterEnd = this->end();
    }
    unsigned int q(RING::my_size());
    unsigned int index;
    unsigned int weight(0);

    for( ; iter<iterEnd; ++iter)
    {
        index = rank % q;
        *iter = Element<RING>(index);
        rank /= q;
        if(index != 0)
        {
            ++weight;
        }
    }
    return weight;
}

template<typename RING>
unsigned int Vec<RING>::getRank(unsigned int len) const
{
    unsigned int res = 0;
    typename Vec<RING>::const_reverse_iterator iter( rbegin() + (this->size() - len ));
    typename Vec<RING>::const_reverse_iterator iterEnd( rend() );
    for(; iter<iterEnd; ++iter)
    {
        res *= RING::my_size();
        res += iter->getIndex();
    }
    return res;
}

template<>
void Vec<Field>::applyFrobenius(unsigned int frob_pow);


template<>
DenseVec<Field> DenseVec<Field>::operator*(const Element<Field>& x) const;

template<>
SymWeight<1u> DenseVec<Field>::getWeight(void) const;

template<>
unsigned int DenseVec<Field>::grayImageWeight(void) const;

// Z4 specific implementations
Element<Z4> DenseVec<Z4>::operator[](unsigned int i) const
{
    unsigned int pos = i % ENTRIES_PER_WORD;
    unsigned int datapos = i / ENTRIES_PER_WORD;
    return Element<Z4> ((m_data[datapos] >> (3 * pos)) & 3);
}

void DenseVec<Z4>::setBits(unsigned int i)
                {
    unsigned int pos = i % ENTRIES_PER_WORD;
    unsigned int datapos = (i - pos) / ENTRIES_PER_WORD;
    m_data[datapos] |= (((uchar) 3) << (3 * pos));
                }

DenseVec<Z4>& DenseVec<Z4>::operator+=(const DenseVec<Z4>& b)
                {
    std::vector<uchar>::iterator iter(m_data.begin());
    std::vector<uchar>::iterator iterEnd(m_data.end());
    std::vector<uchar>::const_iterator iterB(b.m_data.begin());
    for (; iter < iterEnd; ++iter, ++iterB)
    {
        *iter += *iterB;
        *iter &= CLEARMASK;
    }
    return *this;
                }

DenseVec<Z4>& DenseVec<Z4>::operator*=(const Element<Z4>& x)
                {
    if (x.isOne())
    {
        return *this;
    }
    std::vector<uchar>::iterator iter(m_data.begin());
    std::vector<uchar>::iterator iterEnd(m_data.end());
    if (x.isZero())
    {
        for (; iter < iterEnd; ++iter)
        {
            *iter = 0;
        }
    }
    else if (x == 2)
    {
        for (; iter < iterEnd; ++iter)
        {
            *iter <<= 1;
            *iter &= CLEARMASK;
        }
    }
    else
    {
        for (; iter < iterEnd; ++iter)
        {
            *iter += (*iter << 1) & CLEARMASK;
            *iter &= CLEARMASK;
        }
    }
    return *this;
                }

// F2_F2 specific implementation
Element<F2_F2> DenseVec<F2_F2>::operator[](unsigned int i) const
{
    return Element<F2_F2> ((unsigned int) (m_data[i >> 5] >> (2 * (i & DenseVec<F2_F2>::BITPOSMASK))) & 3);
}

void DenseVec<F2_F2>::setBits(unsigned int i)
                {
    m_data[i >> 5] |= ((uchar) 3) << (2 * (i & DenseVec<F2_F2>::BITPOSMASK));
                }

DenseVec<F2_F2>& DenseVec<F2_F2>::operator+=(const DenseVec& b)
                {
    std::vector<uchar>::iterator iter(m_data.begin());
    std::vector<uchar>::iterator iterEnd(m_data.end());
    std::vector<uchar>::const_iterator iterB(b.m_data.begin());
    for (; iter < iterEnd; ++iter, ++iterB)
    {
        *iter ^= *iterB;
    }
    return *this;
                }

DenseVec<F2_F2>& DenseVec<F2_F2>::operator*=(const Element<F2_F2>& x)
                {
    std::vector<uchar>::iterator iter(m_data.begin());
    std::vector<uchar>::iterator iterEnd(m_data.end());
    switch (x.getIndex())
    {
    case 0:
        for (; iter < iterEnd; ++iter)
        {
            *iter = 0;
        }
        return *this;
    case 1:
        return *this;
    case 2:
        for (; iter < iterEnd; ++iter)
        {
            *iter &= DenseVec<F2_F2>::ONEBITMASK; // only units will survive
            *iter <<= 1; // multiplication by X
        }
        return *this;
    case 3:
        for (; iter < iterEnd; ++iter)
        {
            // swap the two bits
            uchar dummy = *iter & DenseVec<F2_F2>::ONEBITMASK;
            *iter &= DenseVec<F2_F2>::ONEBITMASK << 1;
            *iter >>= 1;
            *iter |= dummy;
        }
        return *this;
    }
    return *this;
                }

template<typename RING>
DenseVec<BinaryField>::DenseVec(const Matrix<RING>& m) : m_len(m.getRows() * m.getCols() * RING::bitsPerEntry()), m_data((m_len - 1) / 64 + 1, 0)
{
    const std::vector<Vec<RING> >& cols(m.getEntries());
    typename std::vector<Vec<RING> >::const_iterator colIter(cols.begin());
    typename std::vector<Vec<RING> >::const_iterator colIterEnd(cols.end());

    unsigned int bitposBegin = 0;
    unsigned int bitposEnd = RING::bitsPerEntry();
    unsigned int writepos = 0;

    for (; colIter < colIterEnd; ++colIter)
    {
        typename Vec<RING>::const_iterator colEntIter(colIter->begin());
        typename Vec<RING>::const_iterator colEntIterEnd(colIter->end());

        for (; colEntIter < colEntIterEnd; ++colEntIter)
        {
            if (bitposEnd <= NBITS)
            {
                m_data[writepos] |= ((uchar) colEntIter->getIndex()) << bitposBegin;
                bitposBegin = bitposEnd;
                bitposEnd += RING::bitsPerEntry();
            }
            else
            {
                if (bitposBegin < NBITS)
                {
                    m_data[writepos] |= ((uchar) colEntIter->getIndex()) << bitposBegin; // we loose some bits by shifting...
                }
                ++writepos;
                m_data[writepos] |= ((uchar) colEntIter->getIndex()) >> (NBITS - bitposBegin);

                bitposBegin = RING::bitsPerEntry() - (NBITS - bitposBegin);
                bitposEnd = bitposBegin + RING::bitsPerEntry();
            }
        }
    }
}



#endif /*VECTOR_HPP_*/
